"""
AERoot CLI module (Entry Point)
"""

import argparse
import sys

from aeroot import __version__
from aeroot.aeroot import AERoot, Mode, ProcessNotRunningError, AERootError
from aeroot.util import Logger, error, info, title, EXIT_ERR


def handle_cmd_line() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="AERoot (Android Emulator ROOTing system) v. {}".format(__version__)
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--verbose", "-v", action="store_true", help="show debug level messages")
    group.add_argument("--quiet", "-q", action="store_true", help="quiet output")
    parser.add_argument("--device", "-d", default="emulator-5554", help="specify the device name")
    parser.add_argument("--host", default="127.0.0.1", help="specify adb host")
    parser.add_argument("--port", "-p", type=int, default=5037, help="specify adb port")

    # Mode [pid|name|daemon]
    subparsers = parser.add_subparsers()

    # Search by name mode
    name = subparsers.add_parser("name", help="find a process by name and overwrite credentials")
    name.add_argument("process_name", help="the process name")
    name.set_defaults(mode=Mode.NAME)

    # Search by pid mode
    pid = subparsers.add_parser("pid", help="find a process by PID and overwrite credentials")
    pid.add_argument("pid", type=int, help="the process PID")
    pid.set_defaults(mode=Mode.PID)

    # Daemon mode
    daemon = subparsers.add_parser("daemon", help="overwrite adb daemon credentials")
    daemon.set_defaults(process_name="adbd")
    daemon.set_defaults(mode=Mode.NAME)

    config = parser.parse_args()

    if not hasattr(config, "pid") and not hasattr(config, "process_name"):
        parser.print_help()
        sys.exit(EXIT_ERR)

    return config


def main():
    options = handle_cmd_line()
    Logger.init(options)

    title("AERoot (Android Emulator ROOTing system) v. {}".format(__version__))

    aeroot = AERoot(options)

    try:
        aeroot.do_root()
    except AERootError as err:
        error(f"{err} **Aborting**")
    except ProcessNotRunningError:
        error("Process is not running. Aborting")
    finally:
        aeroot.cleanup()
        info("Exiting")



"""
AERoot utilities module
"""

import argparse
import logging
import sys

try:
    import colorama
    from colorama import Fore, Style
except ModuleNotFoundError:
    pass

EXIT_ERR = 1


class Logger:

    _instance = None


    @staticmethod
    def init(options):
        logging.basicConfig(format="%(message)s")
        logger = logging.getLogger("AEROOT")
        logger.setLevel(Logger.get_log_level(options) if options is not None else logging.INFO)
        Logger._instance = logger
        colorama.init()


    @staticmethod
    def instance():
        return Logger._instance


    @staticmethod
    def get_log_level(config: argparse.Namespace) -> int:
        if config.quiet:
            return logging.CRITICAL

        if config.verbose:
            return logging.DEBUG

        return logging.INFO


def colorize(msg: str, color: str) -> str:
    return "{}{}{}".format(color, msg, Style.RESET_ALL)


def debug(msg: str):
    try:
        prefix = " [{}]".format(colorize("-", Fore.CYAN))
    except NameError:
        prefix = " [-]"

    Logger.instance().debug("%s %s", prefix, msg)


def error(msg: str, do_exit: bool = False):
    try:
        prefix = "[{}]".format(colorize("!", Fore.RED))
    except NameError:
        prefix = "[!]"

    for line in msg.split("\n"):
        Logger.instance().error("%s %s", prefix, line)

    if do_exit:
        sys.exit(EXIT_ERR)


def info(msg: str):
    try:
        prefix = "[{}]".format(colorize("+", Fore.GREEN))
    except NameError:
        prefix = "[+]"

    Logger.instance().info("%s %s", prefix, msg)


def title(msg: str):
    try:
        Logger.instance().info(colorize(msg, Fore.BLUE))
    except NameError:
        Logger.instance().info(msg)



"""
AERoot GDB module
"""

import re
from time import sleep
from typing import List

from pygdbmi.constants import GdbTimeoutError
from pygdbmi.gdbcontroller import GdbController

from .util import info


class GdbError(Exception): pass


class GdbResponse:

    _SPLITTER = "\\t"


    def __init__(self, raw_response):
        self.raw = raw_response


    def _get_payload(self):
        if len(self.raw) < 2:
            return None

        tokens = self.raw[1].get("payload", "").split(GdbResponse._SPLITTER)

        if len(tokens) < 2:
            return None

        return tokens[1].replace("\\n", "")


    def to_int(self):
        payload = self._get_payload()

        return int(payload, 16) if payload is not None else None


    def to_str(self):
        payload = self._get_payload()

        return payload.replace("\\\"", "") if payload is not None else None


class GdbHelper:

    _WRITE_METHODS = {
        1: "write_byte",
        4: "write_dword",
        8: "write_qword"
    }

    _GDB_PY_PATTERN = re.compile(r"--with-python")


    def __init__(self, arch="x86", timeout=180):
        self.arch = arch
        self.timeout = timeout
        self.gdb = None


    def __init_gdb(self):
        if self.gdb is not None:
            return

        self.gdb = GdbController(time_to_check_for_additional_output_sec=self.timeout)

        if self.arch == "x86_64":
            self.gdb.write("set arch i386:x86-64:intel")


    def exit(self):
        if self.gdb is not None:
            try:
                self.gdb.exit()
            except GdbTimeoutError:
                raise GdbError("Can't connect to gdb server")
            finally:
                self.gdb = None


    def start(self):
        self.__init_gdb()

        try:
            self.gdb.write("target remote :1234")
        except GdbTimeoutError:
            raise GdbError("Can't connect to gdb server")


    def stop(self):
        self.gdb.write("detach")


    def write(self, address: int, value: int, size: int = 1):
        getattr(self, GdbHelper._WRITE_METHODS.get(size, "write_byte"))(address, value)


    def write_byte(self, address: int, value: int):
        self.gdb.write("set *(unsigned char*) (%#x) = %#x" % (address, value))


    def write_dword(self, address: int, value: int):
        self.gdb.write("set *(unsigned int*) (%#x) = %#x" % (address, value))


    def write_qword(self, address: int, value: int):
        self.gdb.write("set *(unsigned long*) (%#x) = %#x" % (address, value))


    def read_dword(self, address: int) -> int:
        return GdbResponse(self.gdb.write("x/wx %#x" % address)).to_int()


    def read_addr(self, address: int) -> int:
        return GdbResponse(self.gdb.write("x/a %#x" % address)).to_int()


    def read_str(self, address: int) -> str:
        return GdbResponse(self.gdb.write("x/s %#x" % address)).to_str()


    def read_ip(self) -> int:
        return GdbResponse(self.gdb.write("p/x $pc")).to_int()


    def find(self, query: str) -> List[int]:
        addresses = []
        response = self.gdb.write("find %s" % query, timeout_sec=180)

        for subset in response:
            payload = subset.get("payload")

            if payload is not None and payload.startswith("0x"):
                addresses.append(int(payload.replace("\\n", ""), 16))

        return addresses


    def execute(self, cmd: str) -> filter:
        result = filter(lambda x: x.get("type") == "console" and x.get("payload").startswith("#"),
                        self.gdb.write(cmd))

        return list(result)


    def execute_and_retry(self, cmd, retry_cnt=5, delay=5, msg="Retry"):
        for i in range(1, retry_cnt + 1):
            if i > 1:
                self.stop()
                sleep(delay)
                self.start()

            result = self.execute(cmd)

            if len(result) > 0:
                return result

            info(f"{msg} (try: {i}/{retry_cnt})...")

        return result


    def has_python(self) -> bool:
        try:
            response = self.gdb.write("show configuration")
        except GdbTimeoutError:
            raise GdbError("Can't connect to gdb server")

        return any(map(lambda x: GdbHelper._GDB_PY_PATTERN.search(x) is not None,
                       (r.get("payload", "") for r in response if r.get("type") == "console")))



"""
AERoot AVD module
"""

from argparse import Namespace
from functools import lru_cache
from pathlib import Path

import yaml
from ppadb.client import Client as AdbClient

from aeroot.gdb import GdbHelper, GdbError
from aeroot.util import debug, info

class AmbiguousProcessNameError(Exception): pass
class GdbPythonSupportError(Exception): pass
class AVDError(Exception): pass


class Avd:

    _TASKLIST_CMD = (
        "python",
        "a_swapper = {}",
        "o_tasks = {}",
        "o_pid = {}",

        "addr = gdb.execute('x/a %d'%(a_swapper + o_tasks), to_string=True).split(':\\t')[1]",
        "addr = int(addr, 16) - o_tasks",

        "while addr != a_swapper:",
        "   pid = gdb.execute('x/wx %d'%(addr + o_pid), to_string=True).split(':\\t')[1]",
        "   pid = int(pid.replace('\\n', ''), 16)",
        "   print('#%d;%d'%(addr, pid))",
        "   addr = gdb.execute('x/a %d'%(addr + o_tasks), to_string=True).split(':\\t')[1]",
        "   addr = int(addr, 16) - o_tasks",
        "end"
    )

    _CAPABILITIES_OFFSETS = [0x30, 0x34, 0x38, 0x3c, 0x40, 0x44]
    _IDS_OFFSETS = [0x04, 0x08, 0x0c, 0x10, 0x14, 0x18, 0x1c, 0x20, 0x24]


    def __init__(self, device: str, host: str, port: int):
        self._tasklist = None

        try:
            self.device = AdbClient(host=host, port=port).device(device)
        except RuntimeError as err:
            raise AVDError(err)

        if self.device is None:
            raise AVDError("Can't connect to emulator through ADB")


    @property
    @lru_cache(maxsize=1)
    def kernel(self):
        config_name = "{}.yaml".format(self.device.shell("uname -rm").replace(" ", "_").strip())
        root_path = Path(__file__).resolve().parent.parent

        return Kernel.load(root_path / "config" / "kernel" / config_name)


    @property
    def tasklist(self):
        if self._tasklist is not None:
            return self._tasklist

        debug("Retrieving tasklist from memory")

        cmd = "\n".join(Avd._TASKLIST_CMD).format(self.kernel.swapper_address,
                                                  self.kernel.config.task.offset.tasklist,
                                                  self.kernel.config.task.offset.pid)

        try:
            results = self.kernel.gdb.execute_and_retry(cmd, msg="Wait for kernel memory mapping")
        except GdbError as err:
            raise AVDError(err)

        tasklist = dict()

        for result in results:
            addr, pid = result.get("payload").replace("\\n", "").replace("#", "", 1).split(";")
            tasklist[int(pid)] = int(addr)

        self._tasklist = tasklist if len(tasklist) > 0 else None
        return self._tasklist


    def find_process(self, pid: int):
        info(f"Kernel base address found at 0x{self.kernel.base_address:x}")

        tasklist = self.tasklist

        if tasklist is None:
            raise AVDError("Can't retrieve tasklist from emulator memory")

        paddr = tasklist.get(int(pid))
        info(f"Process [{pid}] found at 0x{paddr:x}")

        return paddr


    def get_pid(self, pname: str) -> int:
        pids = self.device.shell(f"pidof {pname}").replace("\\n", "").split()

        if len(pids) > 1:
            raise AmbiguousProcessNameError()

        return int(pids[0]) if len(pids) > 0 else None


    def overwrite_credentials(self, pid):
        address = self.find_process(pid) + self.kernel.config.task.offset.creds
        cmd = [f"x/a {address}", "set $addr = $__"]

        for offset in Avd._CAPABILITIES_OFFSETS:
            cmd.append("set *(unsigned int*) ($addr + {}) = {}".format(offset, 0xffffffff))

        for offset in Avd._IDS_OFFSETS:
            cmd.append("set *(unsigned int*) ($addr + {}) = {}".format(offset, 0x00000000))

        info(f"Overwriting process [{pid}] credentials")

        self.kernel.gdb.execute("\n".join(cmd))


    def selinux_setenforce(self, mode: int):
        self.kernel.enforce = mode


    def close(self):
        try:
            self.kernel.gdb.exit()
        except AVDError:
            pass


class Kernel:

    _KERNEL_BASE_CMD = (
        "python",
        "range_start = {}",
        "range_stop = {}",
        "found = False",

        "for addr in range(range_start, range_stop, 0x1000000):",
        "   if found: break",

        "   try:",
        "       gdb.execute('x/a %d'%addr, to_string=True)",
        "       k_addr = addr",
        "       c_addr = addr - 0x100000",
        "       while c_addr > addr - 0x1000000:",
        "           try:",
        "               gdb.execute('x/a %d'%c_addr, to_string=True)",
        "               k_addr = c_addr",
        "               c_addr -= 0x100000",
        "           except gdb.MemoryError:",
        "               print('#%d'%k_addr)",
        "               found = True",
        "               break",
        "   except gdb.MemoryError:",
        "       pass",
        "end"
    )


    def __init__(self, config):
        self.config = config
        self._base_address = None


    @property
    @lru_cache(maxsize=1)
    def gdb(self):
        gdb = GdbHelper(arch=self.config.arch)

        try:
            gdb.start()

            if not gdb.has_python():
                raise GdbPythonSupportError
        except GdbError as err:
            raise AVDError(err)

        return gdb


    @property
    def base_address(self):
        if self._base_address is not None:
            return self._base_address

        debug("Retrieving kernel base address from memory")

        cmd = "\n".join(Kernel._KERNEL_BASE_CMD).format(self.config.mem_range.begin,
                                                        self.config.mem_range.end)

        try:
            result = self.gdb.execute_and_retry(cmd, msg="Wait for kernel memory mapping")
        except GdbError as err:
            raise AVDError(err)

        if len(result) == 0:
            raise AVDError("Can't retrieve kernel base from memory")

        self._base_address = int(result[0].get("payload").replace("#", "").replace("\\n", ""))
        return self._base_address


    @property
    @lru_cache(maxsize=1)
    def selinux_address(self):
        return self.base_address + self.config.offset.selinux


    @property
    @lru_cache(maxsize=1)
    def swapper_address(self):
        return self.base_address + self.config.offset.swapper


    @property
    def enforce(self):
        return self.gdb.read_dword(self.selinux_address)


    @enforce.setter
    def enforce(self, mode):
        info(f"Set SELinux enforce (0x{self.selinux_address:x}) to {mode}")
        self.gdb.write(self.selinux_address, mode, self.config.sizeof.enforce)


    @staticmethod
    def load(filename):
        with open(Path(Path.cwd(), "config", "kernel", filename), "r") as fconfig:
            config = Kernel._get_config(yaml.load(fconfig, yaml.FullLoader))

            return Kernel(config)


    @staticmethod
    def _get_config(mapping):
        if isinstance(mapping, dict):
            for key, value in mapping.items():
                mapping[key] = Kernel._get_config(value)
            return Namespace(**mapping)

        return mapping



#!/usr/bin/env python3

from aeroot import __version__
from os import path
from setuptools import setup, find_packages

main_dir = path.abspath(path.dirname(__file__))
install_requires = open(path.join(main_dir, "requirements.txt"), "r").readlines()
long_description = open(path.join(main_dir, "README.md"), "r", encoding="utf-8").read()
packages = find_packages(exclude=["*.tests", "*.tests.*", "test*", "tests"])
packages.append("config")

setup(
    name="aeroot",
    version=__version__,
    packages=packages,
    url="https://github.com/quarkslab/AERoot",
    license="Apache License 2.0",
    author="Eric Le Guevel (ha0)",
    author_email="eleguevel@quarkslab.com",
    description="Android Emulator Rooting system",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=install_requires,
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "aeroot=aeroot.cli:main",
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Security",
        "Topic :: Utilities",
    ]
)
